#ifndef GAMEFIELD_H
#define GAMEFIELD_H

#include <SFML/Graphics.hpp>
#include <set>

class Gamefield
{
    public:
        static const int WIDTH = 7;
        static const int HEIGHT = 12;
        static const int CELLSIZE = 64;

        Gamefield();
        virtual ~Gamefield();

        void RenderToConsole();
        void RenderToTarget(sf::RenderTarget& target);

        const bool Init();
        void Reset();
        void Start();

        // called once per timestep (not per frame!)
        void Update();

        //player interaction
        void MoveLeft();
        void MoveRight();

    protected:
    private:
        static const int MINMATCH = 3; // number of same-colored balls required for a match
        static const float MULTIPLIERFACTOR = 0.25f; //by how much the multiplier increases per additional match > MINMATCH
        static const int MULTIPLIERTIME = 5; //how many rounds the multiplier remains
        static const int SPRINGHEIGHT = 6; // how long are springs on default?
        static const int MINSPRINGHEIGHT = 2; // how short can springs be?
        static const sf::Color COLORS[];
        static const int NUMCOLORS_; //had a nameclash here
        static const int STARTCOLORS = 3;
        static const int NEXTCOLORSCORE[]; //get a new color once you reach n points

        int mMultiplierCountdown;

        int mCurColors;

        char mMap[HEIGHT][WIDTH]; // A...Z = falling ball, a..z = non-falling ball, # = spring, ' ' = empty

        sf::Image mImageBall;
        sf::Image mImageSpring;
        sf::Image mImageHolder;
        sf::Sprite mSprite;

        sf::Vector2i mPlayerBallPosition;
        bool mCreateNewPlayerBall; //whether a new player ball should be created

        //parts of update:
        //1.: let every falling block fall further
        void LetThemFall();
        //2.: check springs: do they need to be moved due to changed load?
        void CheckSprings();
        //3.: remove matching balls, award points, set the others to "falling"
        void RemoveMatches();

        int GetSpringSize(const int x);

        struct VectorSetComparator
        {
            bool operator()(const sf::Vector2i& lhs, const sf::Vector2i& rhs) const
            {
                return lhs.x * Gamefield::HEIGHT + lhs.y < rhs.x * Gamefield::HEIGHT + rhs.y;
            }
        };

        typedef std::set<sf::Vector2i, VectorSetComparator> VectorSet;

        //recursive function for finding all the same colored neighbours
        void FindMatchingNeighbours(int x, int y, char desired, VectorSet& matches);

        void CreatePlayerBall();

        void Score(int amount);
};

#endif // GAMEFIELD_H
