#include "Gamefield.h"
#include <string>
#include <iostream>
#include <cassert>
#include <algorithm>

const sf::Color Gamefield::COLORS[] = //if you add more, remember to also add more NEXTCOLORSCOREs
{
    sf::Color(255, 100, 100),
    sf::Color(100, 255, 100),
    sf::Color(100, 100, 255),
    sf::Color(100, 0, 150),
    sf::Color(0, 255, 255),
    sf::Color(255, 255, 255),
    sf::Color(255, 255, 100),
    sf::Color(50, 50, 50),
};
const int Gamefield::NUMCOLORS_ = sizeof(COLORS) / sizeof(COLORS[0]);
const int Gamefield::NEXTCOLORSCORE[] =
{
    15,
    50,
    120,
    200,
    300
};

extern bool g_youLostTheGame;
extern int g_score;
extern float g_multiplier;

Gamefield::Gamefield() :
    mPlayerBallPosition(-1, -1),
    mCreateNewPlayerBall(false)
{
    //ctor
    for(int y = 0; y < HEIGHT; ++y)
    {
        for(int x = 0; x < WIDTH; ++x)
        {
            mMap[y][x] = ' ';
        }
    }
}

const bool Gamefield::Init()
{
    if(!mImageBall.LoadFromFile("data/ball.png")) return false;
    if(!mImageSpring.LoadFromFile("data/spring.png")) return false;
    if(!mImageHolder.LoadFromFile("data/holder.png")) return false;
    Reset();
    return true;
}

void Gamefield::Reset()
{
    for(int y = 0; y < HEIGHT; ++y)
    {
        for(int x = 0; x < WIDTH; ++x)
        {
            if(y < Gamefield::SPRINGHEIGHT)
            {
                mMap[y][x] = '#';
            }
            else
            {
                mMap[y][x] = ' ';
            }
        }
    }

    mPlayerBallPosition = sf::Vector2i(-1, -1);
    mCurColors = STARTCOLORS;
    mMultiplierCountdown = 0;
}

Gamefield::~Gamefield()
{
    //dtor
}

void Gamefield::Start()
{
    mCreateNewPlayerBall = true;
}

void Gamefield::RenderToConsole()
{
    std::string output = "\n\n\n\n\n\n\n\n\n+";
    for(int x = 0; x < WIDTH; ++x) output += '-';
    output += '+';
    for(int y = HEIGHT - 1; y >= 0; --y)
    {
        output += "\n|";
        for(int x = 0; x < WIDTH; ++x)
        {
            output += mMap[y][x];
        }
        output += '|';
    }
    output += "\n+";
    for(int x = 0; x < WIDTH; ++x) output += '-';
    output += '+';
    std::cout<<output<<std::endl;
}


int Gamefield::GetSpringSize(const int x)
{
    int springsize = 0;
    while(mMap[springsize++][x] == '#');
    return springsize - 1;
}

void Gamefield::RenderToTarget(sf::RenderTarget& target)
{
    static bool first = true;
    for(int x = 0; x < WIDTH; ++x)
    {
        //  render spring

        //figure its total size out
        int springsize = GetSpringSize(x);

        //figure the height of a single element out
        float height = (float)(springsize) / (SPRINGHEIGHT) * CELLSIZE;

        mSprite.SetColor(sf::Color::White);
        mSprite.SetImage(mImageSpring);
        mSprite.Resize(CELLSIZE, height+1);
        for(int y = 0; y < SPRINGHEIGHT; ++y)
        {
            mSprite.SetPosition(x*CELLSIZE, 64+(HEIGHT)*CELLSIZE - (y+1) * height);
            target.Draw(mSprite);
        }

        mSprite.SetImage(mImageHolder);
        mSprite.Resize(CELLSIZE, CELLSIZE);
        mSprite.SetPosition(x*CELLSIZE, 64+((float)HEIGHT - springsize - 0.5f) * CELLSIZE);
        target.Draw(mSprite);

        //render balls
        for(int y = 0; y < HEIGHT; ++y)
        {
            bool draw = false;
            if(mMap[y][x] >= 'a' && mMap[y][x] <= 'a' + sizeof(COLORS) - 1)
            {
                mSprite.SetColor(COLORS[mMap[y][x]-'a']);
                draw = true;
            }
            if(mMap[y][x] >= 'A' && mMap[y][x] <= 'A' + sizeof(COLORS) - 1)
            {
                mSprite.SetColor(COLORS[mMap[y][x]-'A']);
                draw = true;
            }

            if(draw)
            {
                mSprite.SetImage(mImageBall);
                mSprite.Resize(CELLSIZE, CELLSIZE);
                mSprite.SetPosition(x*CELLSIZE, 64+(HEIGHT-y-1)*CELLSIZE);
                target.Draw(mSprite);
            }
        }
    }
    first = false;
}

void Gamefield::Update()
{
#ifdef _DEBUG
    std::cout<<"Update!"<<std::endl;
#endif
    if(g_youLostTheGame) return;
    LetThemFall();
    CheckSprings();
    RemoveMatches();
    if(mCreateNewPlayerBall) CreatePlayerBall();
}

static bool IsNotFallingBall(char c)
{
    return(c >= 'a' && c <= 'z');
}

static bool IsFalling(char c)
{
    return(c >= 'A' && c <= 'Z'); //A-Z = falling
}

void Gamefield::LetThemFall()
{
    int difLower = 'A' - 'a';
    for(int x = 0 ; x < WIDTH ; ++x)
    {
        //from bottom to top so it's not delayed, starting with 1 because 0 can't fall down.
        for(int y = 1 ; y < HEIGHT ; ++y)
        {
            //is this a falling ball?
            if(IsFalling(mMap[y][x]))
            {
                bool isPlayerBall = sf::Vector2i(x, y) == mPlayerBallPosition;
                //is there space below it?
                if(mMap[y-1][x] == ' ')
                {
                    //yes there is. Move it down, not making non-falling (yet) since we want blocks to be there for one tick before removing matches

                    mMap[y-1][x] = mMap[y][x];
                    if(isPlayerBall) --mPlayerBallPosition.y;

                    //at first I forgot this: remove it from its old position^^
                    mMap[y][x] = ' ';
                }
                else
                {
                    //no space below it, it's not falling anymore!
                    mMap[y][x] = mMap[y][x] - difLower;
                    //so the player can't control it anymore, either.
                    if(isPlayerBall)
                    {
                        mPlayerBallPosition = sf::Vector2i(-1, -1);
                        //he needs a new ball.
                        mCreateNewPlayerBall = true;
                    }
                }
            }
        }
    }
}

static char MakeFalling(char c)
{
    if(c > 'z' || c < 'a') return c;
    return c - 'a' + 'A';
}

void Gamefield::CheckSprings()
{
    for(int x = 0 ; x < WIDTH ; ++x)
    {
        int springsize = GetSpringSize(x);
        int dif = SPRINGHEIGHT - springsize;
        assert(dif >= 0); //springs should never be longer than at the beginning!

        int numBallsPush = 0; //falling balls are not counted for pushing it down
        int numBallsPull = 0; //but they are counted for moving back up
        for(int y = springsize; y < HEIGHT; ++y)
        {
            assert(mMap[y][x] != '#');
            if(mMap[y][x] == ' ')
            {
                break;
            }
            else
            {
                if(!IsFalling(mMap[y][x])) ++numBallsPush;
                ++numBallsPull;
            }
        }
        if(numBallsPull < dif) //not enough weight
        {
            //move all up by one, unless we're already at the top - that's a loss.
            if(numBallsPull + springsize >= HEIGHT)
            {
                g_youLostTheGame = true;
                return;
            }
            for(int y = numBallsPull + springsize; y >= springsize; --y)
            {
                if(y == 0) //special case, won't happen with the default values
                {
                    mMap[y][x] = '#';
                }
                else
                {
                    //make them falling again so we've got one tick until matches are removed.
                    mMap[y][x] = MakeFalling(mMap[y-1][x]);
                    if(sf::Vector2i(x, y-1) == mPlayerBallPosition) mPlayerBallPosition.y = y;
                }
            }
        }
        //the !IsFalling part is to prevent the spring from lowering at the tick when it first receives weight. could this cause bugs in other situations?
        else if(numBallsPush > dif && springsize > MINSPRINGHEIGHT && !IsFalling(mMap[springsize][x])) //too much weight
        {
            //move all down by one
            for(int y = springsize - 1; y < springsize + numBallsPush; ++y)
            {
                if(y < 0) //special case, won't happen with the default values
                {
                    continue;
                }
                else
                {
                    //make them falling again so we've got one tick until matches are removed.
                    if(y+1 < HEIGHT)
                    {
                        mMap[y][x] = MakeFalling(mMap[y+1][x]);
                        if(sf::Vector2i(x, y+1) == mPlayerBallPosition) mPlayerBallPosition.y = y;
                    }
                    else
                    {
                        mMap[y][x] = ' ';
                    }
                }
            }
        }
    }
}

void Gamefield::RemoveMatches()
{
    int addScore=0;
    for(int x = 0 ; x < WIDTH ; ++x)
    {
        for(int y = 0; y < HEIGHT ; ++y)
        {
            if(IsNotFallingBall(mMap[y][x]))
            {
                VectorSet matches; //so I don't count blocks twice
                FindMatchingNeighbours(x, y, mMap[y][x], matches);
                //enough of the same color?
                if(matches.size() >= MINMATCH)
                {
                    //remove them
                    for(VectorSet::iterator it = matches.begin(); it != matches.end(); ++it)
                    {
                        mMap[it->y][it->x] = ' ';
                    }
                    addScore += matches.size();

                }
            }
        }
    }

    if(mMultiplierCountdown > 0)
    {
        --mMultiplierCountdown;
    }
    float multiplierChange = MULTIPLIERFACTOR * (addScore - MINMATCH);
    if(multiplierChange >= 0)
    {
        g_multiplier += multiplierChange;
        mMultiplierCountdown = MULTIPLIERTIME;
    }
    else if(mMultiplierCountdown <= 0)
    {
        g_multiplier = std::max(1.0f, g_multiplier - MULTIPLIERFACTOR);
    }
    Score(addScore * g_multiplier);
    //now that I removed some, many blocks must be falling...
    for(int x = 0 ; x < WIDTH; ++x)
    {
        bool makeRestFall = false;
        //looking from bottom to top - once I find an empty space the rest must be falling.
        for(int y = 0 ; y < HEIGHT ; ++y)
        {
            if(mMap[y][x] == ' ') makeRestFall = true;
            else if(makeRestFall) mMap[y][x] = MakeFalling(mMap[y][x]);
        }
    }
}

void Gamefield::FindMatchingNeighbours(int x, int y, char desired, Gamefield::VectorSet& matches)
{
    if(mMap[y][x] != desired) return; //not the right color?

    if(matches.insert(sf::Vector2i(x, y)).second == false) return; //already existed?

    //check neighbours
    if(x+1 < WIDTH) FindMatchingNeighbours(x+1, y, desired, matches);
    if(x-1 >= 0) FindMatchingNeighbours(x-1, y, desired, matches);
    //if(y+1 < HEIGHT) FindMatchingNeighbours(x, y+1, desired, matches);
    //if(y-1 >= 0) FindMatchingNeighbours(x, y-1, desired, matches);
}

void Gamefield::CreatePlayerBall()
{
    assert(mPlayerBallPosition == sf::Vector2i(-1, -1));
    mCreateNewPlayerBall = false;
    mPlayerBallPosition.x = WIDTH / 2;
    mPlayerBallPosition.y = HEIGHT - 1;
    if(mMap[mPlayerBallPosition.y][mPlayerBallPosition.x] != ' ')
    {
        g_youLostTheGame = true;
        return;
    }
    mMap[mPlayerBallPosition.y][mPlayerBallPosition.x] = 'A' + sf::Randomizer::Random(0, mCurColors-1);
}

void Gamefield::MoveLeft()
{
    if(mPlayerBallPosition == sf::Vector2i(-1, -1)) return; //no ball available
    if(mPlayerBallPosition.x <= 0) return; //already as left as possible
    if(mMap[mPlayerBallPosition.y][mPlayerBallPosition.x-1] != ' ') return; //no space
    mMap[mPlayerBallPosition.y][mPlayerBallPosition.x-1] = mMap[mPlayerBallPosition.y][mPlayerBallPosition.x];
    mMap[mPlayerBallPosition.y][mPlayerBallPosition.x] = ' ';
    --mPlayerBallPosition.x;
}

void Gamefield::MoveRight()
{
    if(mPlayerBallPosition == sf::Vector2i(-1, -1)) return; //no ball available
    if(mPlayerBallPosition.x >= WIDTH - 1) return; //already as right as possible
    if(mMap[mPlayerBallPosition.y][mPlayerBallPosition.x+1] != ' ') return; //no space
    mMap[mPlayerBallPosition.y][mPlayerBallPosition.x+1] = mMap[mPlayerBallPosition.y][mPlayerBallPosition.x];
    mMap[mPlayerBallPosition.y][mPlayerBallPosition.x] = ' ';
    ++mPlayerBallPosition.x;
}

void Gamefield::Score(int amount)
{
    g_score += amount;
    if(mCurColors == NUMCOLORS_) return; //already at last color

    if(g_score >= NEXTCOLORSCORE[mCurColors - STARTCOLORS]) ++mCurColors; //enough points for next color
}
